@@ -1,7 +1,7 @@
 /*
  Project: FTP
 
- Copyright (C) 2005 Riccardo Mottola
+ Copyright (C) 2005-2015 Riccardo Mottola
 
  Author: Riccardo Mottola
 
@@ -28,12 +28,40 @@
 #import <AppKit/AppKit.h>
 #import "fileElement.h"
 
+#define TAG_FILENAME @"filename"
 
-@interface fileTable : NSObject {
-    NSArray        *fileStructs;
+#if !defined (GNUSTEP) &&  (MAC_OS_X_VERSION_MAX_ALLOWED <= MAC_OS_X_VERSION_10_4)
+#ifndef NSInteger
+#define NSInteger int
+#endif
+#ifndef NSUInteger
+#define NSUInteger unsigned
+#endif
+#endif
+
+
+enum sortOrderDef
+{
+  ascending, descending, undefined
+};
+
+@interface FileTable : NSObject
+{
+  NSMutableArray *fileStructs;
+  NSMutableArray *sortedArray;
+  NSString       *sortByIdent;
+  enum sortOrderDef   sortOrder;
 }
 
 - (void)initData:(NSArray *)fnames;
-- (fileElement *)elementAtIndex:(unsigned)index;
+- (void)clear;
+- (FileElement *)elementAtIndex:(NSUInteger)index;
+- (void)generateSortedArray;
+- (void)addObject:(FileElement *)object;
+- (void)removeObjectAtIndex:(NSUInteger)index;
+- (void)removeObject:(FileElement *)object;
+- (void)sortByIdent:(NSString *)idStr;
+- (BOOL)containsFileName:(NSString *)name;
+
 
 @end
