@@ -1,7 +1,7 @@
 /*
  Project: FTP
 
- Copyright (C) 2005-2007 Riccardo Mottola
+ Copyright (C) 2005-2015 Riccardo Mottola
 
  Author: Riccardo Mottola
 
@@ -55,7 +55,7 @@
 #define MAX_CONTROL_BUFF 2048
 #define MAX_DATA_BUFF 2048
 
-#if defined(__linux__)
+#if defined(__linux__) || defined (__BSD_VISIBLE)
 #define socklentype socklen_t
 #else
 #define socklentype int
@@ -63,50 +63,55 @@
 
 void initStream(streamStruct *ss, int socket)
 {
-    ss->socket = socket;
-    ss->position = 0;
-    ss->len = 0;
-    ss->buffer[0] = '\0';
+  ss->socket = socket;
+  ss->position = 0;
+  ss->len = 0;
+  ss->buffer[0] = '\0';
 }
 
 int getChar(streamStruct* ss)
 {
-    int result;
-    BOOL gotEof;
+  int result;
+  BOOL gotEof;
 
-    gotEof = NO;
-    if (ss->position == ss->len)
+  gotEof = NO;
+  if (ss->position == ss->len)
     {
-        int read;
+      int read;
 
-        read = recv(ss->socket, ss->buffer, MAX_SOCK_BUFF, 0);
-        if (read > 0)
+      read = recv(ss->socket, ss->buffer, MAX_SOCK_BUFF, 0);
+      if (read > 0)
         {
-            ss->len = read;
-            ss->position = 0;
-        } else if (read == 0)
+          ss->len = read;
+          ss->position = 0;
+        }
+      else if (read == 0)
         {
-            ss->len = 0;
-            ss->position = 0;
-            ss->buffer[0] = '\0';
-            gotEof = YES;
-        } else
+          ss->len = 0;
+          ss->position = 0;
+          ss->buffer[0] = '\0';
+          gotEof = YES;
+        } 
+      else
         {
-            ss->len = 0;
-            ss->position = 0;
-            NSLog(@"error sock read");
-            perror("getChar:read");
-            ss->buffer[0] = '\0';
-            gotEof = YES;
+          ss->len = 0;
+          ss->position = 0;
+          NSLog(@"error sock read");
+          perror("getChar:read");
+          ss->buffer[0] = '\0';
+          gotEof = YES;
         }
     }
-    if (gotEof)
-        result = EOF;
-    else {    
-        result = ss->buffer[ss->position];
-        ss->position++;
+  if (gotEof)
+    {
+      result = EOF;
     }
-    return result;
+  else
+    {    
+      result = ss->buffer[ss->position];
+      ss->position++;
+    }
+  return result;
 }
 
 @implementation FtpClient
@@ -119,9 +124,8 @@ int getChar(streamStruct* ss)
 	
     pool = [[NSAutoreleasePool alloc] init];
 	
-    serverConnection = [NSConnection
-            connectionWithReceivePort:[portArray objectAtIndex:0]
-							 sendPort:[portArray objectAtIndex:1]];
+    serverConnection = [NSConnection connectionWithReceivePort: [portArray objectAtIndex:0]
+                                                      sendPort:[portArray objectAtIndex:1]];
 	
     serverObject = [self alloc];
     [(id)[serverConnection rootProxy] setServer:serverObject];
@@ -158,7 +162,6 @@ int getChar(streamStruct* ss)
     wVersionRequested = MAKEWORD( 1, 1 );
 
     WSAStartup(wVersionRequested, &wsaData);
-    NSLog(@"inited WinSock");
 #endif
     connected = NO;
     return self;
@@ -189,20 +192,18 @@ int getChar(streamStruct* ss)
  */
 - (void)changeWorkingDir:(NSString *)dir
 {
-    char            tempStr[MAX_CONTROL_BUFF];
-    char            tempStr2[MAX_CONTROL_BUFF];
-    NSMutableArray *reply;
+  NSString       *tempStr;
+  NSMutableArray *reply;
 
-    if (!connected)
-        return;
-	
-    [dir getCString:tempStr2];
-    sprintf(tempStr, "CWD %s\r\n", tempStr2);
-    [self writeLine:tempStr];
-    if ([self readReply:&reply] == 250)
-        [super changeWorkingDir:dir];
-    else
-        NSLog(@"cwd failed");
+  if (!connected)
+    return;
+
+  tempStr = [@"CWD " stringByAppendingString:dir];
+  [self writeLine:tempStr];
+  if ([self readReply:&reply] == 250)
+    [super changeWorkingDir:dir];
+  else
+    NSLog(@"cwd failed");
 }
 
 /* if we have a valid controller, we suppose it respons to appendTextToLog */
@@ -246,13 +247,14 @@ int getChar(streamStruct* ss)
     state = N1;
     separator = 0;
     multiline = NO;
+    startNumCode = numCode = 0;
     *result = [NSMutableArray arrayWithCapacity:1];
 
     // TODO: protect against numCodeStr overflow
     while (!(state == END))
     {
         ch = getChar(&ctrlStream);
-//	NSLog(@"read char: %c", ch);
+	//NSLog(@"read char: %c", ch);
         if (ch == EOF)
             state = END;
 
@@ -324,6 +326,7 @@ int getChar(streamStruct* ss)
                 break;
             case END:
                 NSLog(@"EOF reached prematurely");
+		startNumCode = -1;
                 break;
             default:
                 NSLog(@"Duh, a case default in the readReply parser");
@@ -336,101 +339,99 @@ int getChar(streamStruct* ss)
 /*
  writes a single line to the control connection, logging it always
  */
-- (int)writeLine:(char *)line
+- (int)writeLine:(NSString *)line
 {
-    return [self writeLine:line byLoggingIt:YES];
+  return [self writeLine:line byLoggingIt:YES];
 }
 
 /*
  writes a single line to the control connection
  */
-- (int)writeLine:(char *)line byLoggingIt:(BOOL)doLog
+- (int)writeLine:(NSString *)line byLoggingIt:(BOOL)doLog
 {
-    int sentBytes;
-    int bytesToSend;
+  int  sentBytes;
+  int  bytesToSend;
+  char command[MAX_CONTROL_BUFF];
+  NSString *commandStr;
 
-    bytesToSend = strlen(line);
-    if (doLog)
-        [self logIt:[NSString stringWithCString:line length:(bytesToSend - 2)]];
-    if ((sentBytes = send(controlSocket, line, strlen(line), 0)) < bytesToSend)
-        NSLog(@"sent %d out of %d", sentBytes, bytesToSend);
-    return sentBytes;
+  commandStr = [line stringByAppendingString:@"\r\n"];
+  [commandStr getCString:command];
+  bytesToSend = strlen(command);
+  if (doLog)
+    [self logIt:line];
+  if ((sentBytes = send(controlSocket, command, bytesToSend, 0)) < bytesToSend)
+    NSLog(@"sent %d out of %d", sentBytes, bytesToSend);
+  return sentBytes;
 }
 
 
 - (int)setTypeToI
 {
-    NSMutableArray *reply;
-    int            retVal;
-    
-    retVal = [self writeLine:"TYPE I\r\n"];
-    if ( retVal > 0)
+  NSMutableArray *reply;
+  int            retVal;
+  
+  retVal = [self writeLine:@"TYPE I"];
+  if (retVal > 0)
     {
-        [self readReply:&reply];
-        [reply release];
+      [self readReply:&reply];
+      [reply release];
     }
-       
-    return retVal;
+  
+  return retVal;
 }
 
 - (int)setTypeToA
 {
-    NSMutableArray *reply;
-    int            retVal;
-
-    retVal = [self writeLine:"TYPE A\r\n"];
-    NSLog(@"retval: %d", retVal);
-    if ( retVal > 0)
+  NSMutableArray *reply;
+  int            retVal;
+  
+  retVal = [self writeLine:@"TYPE A"];
+  if (retVal > 0)
     {
-        [self readReply:&reply];
-        [reply release];
+      [self readReply:&reply];
+      [reply release];
     }
-    
-    return retVal;
+  
+  return retVal;
 }
 
-- (oneway void)retrieveFile:(fileElement *)file to:(LocalClient *)localClient beingAt:(int)depth;
+- (BOOL)retrieveFile:(FileElement *)file to:(LocalClient *)localClient beingAt:(int)depth;
 {
-    NSString           *fileName;
-    unsigned long long fileSize;
-    char               fNameCStr[MAX_CONTROL_BUFF];
-    char               command[MAX_CONTROL_BUFF];
-    char               buff[MAX_DATA_BUFF];
-    FILE               *localFileStream;
-    int                bytesRead;
-    NSMutableArray     *reply;
-    struct sockaddr    from;
-    int                fromLen;
-    unsigned int       minimumPercentIncrement;
-    unsigned int       progressIncBytes;
-    int                replyCode;
-    unsigned long long totalBytes;
-    NSString           *localPath;
-    BOOL               gotFile;
+  NSString           *fileName;
+  unsigned long long fileSize;
+  NSString           *command;
+  char               buff[MAX_DATA_BUFF];
+  FILE               *localFileStream;
+  int                bytesRead;
+  NSMutableArray     *reply;
+  unsigned int       minimumPercentIncrement;
+  unsigned int       progressIncBytes;
+  int                replyCode;
+  unsigned long long totalBytes;
+  NSString           *localPath;
+  BOOL               gotFile;
 
-    fromLen = sizeof(from);    
+  fileName = [file name];
+  fileSize = [file size];
+  minimumPercentIncrement = fileSize / 100; // we should guard against maxint
 
-    fileName = [file filename];
-    fileSize = [file size];
-    minimumPercentIncrement = fileSize / 100; // we should guard against maxint
+  localPath = [[localClient workingDir] stringByAppendingPathComponent:fileName];
 
-    localPath = [[localClient workingDir] stringByAppendingPathComponent:fileName];
-
-    if ([file isDir])
+  if ([file isDir])
     {
-        NSString     *pristineLocalPath;  /* original path */
-        NSString     *pristineRemotePath; /* original path */
-        NSArray      *dirList;
-        NSString     *remoteDir;
-        NSEnumerator *en;
-        fileElement  *fEl;
+      NSString     *pristineLocalPath;  /* original path */
+      NSString     *pristineRemotePath; /* original path */
+      NSArray      *dirList;
+      NSString     *remoteDir;
+      NSEnumerator *en;
+      FileElement  *fEl;
 
-        if (depth > MAX_DIR_RECURSION)
+      if (depth > MAX_DIR_RECURSION)
         {
-            NSLog(@"Max depth reached: %d", depth);
-            return;
+          NSLog(@"Max depth reached: %d", depth);
+          return NO;
         }
-
+      
         pristineLocalPath = [[localClient workingDir] retain];
         pristineRemotePath = [[self workingDir] retain];
         
@@ -438,23 +439,23 @@ int getChar(streamStruct* ss)
         [self changeWorkingDir:remoteDir];
 
         if ([localClient createNewDir:localPath] == YES)
-        {
+          {
             [localClient changeWorkingDir:localPath];
     
             dirList = [self dirContents];
             en = [dirList objectEnumerator];
             while ((fEl = [en nextObject]))
-            {
-                NSLog(@"recurse, download : %@", [fEl filename]);
+              {
+                NSLog(@"recurse, download : %@", [fEl name]);
                 [self retrieveFile:fEl to:localClient beingAt:depth+1];
-            }
-        }
+              }
+          }
         /* we get back were we started */
         [self changeWorkingDir:pristineRemotePath];
         [localClient changeWorkingDir:pristineLocalPath];
         [pristineLocalPath release];
         [pristineRemotePath release];
-        return;
+        return YES;
     }
 
     /* lets settle to a plain binary standard type */
@@ -463,27 +464,61 @@ int getChar(streamStruct* ss)
     if ([self initDataConn] < 0)
     {
         NSLog(@"error initiating data connection, retrieveFile");
-        return;
+        return NO;
     }
-    
-    [fileName getCString:fNameCStr];
-    sprintf(command, "RETR %s\r\n", fNameCStr);
+
+    command = [@"RETR " stringByAppendingString:fileName];
     [self writeLine:command];
     replyCode = [self readReply:&reply];
     NSLog(@"%d reply is %@: ", replyCode, [reply objectAtIndex:0]);
 
     if(replyCode != 150)
-    {
+      {
         [controller showAlertDialog:@"Unexpected server error."];
         NSLog(@"Unexpected condition in retrieve");
-        return; /* we have an error or some unexpected condition */
-    }
+        return NO; /* we have an error or some unexpected condition */
+      }
+    else
+      {
+	NSString *s;
+	NSRange bytesR;
+
+	/* we try to parse the response which may look like this:
+	   150 Opening BINARY mode data connection for core.current.tar.bz2 (10867411 bytes)
+	   and extract the transfer size */
+	s = [reply objectAtIndex:0];
+	bytesR = [s rangeOfString:@")" options:NSBackwardsSearch];
+	if (bytesR.location > 0)
+	  {
+	    NSRange leftParR;
+	    NSString *sizeString;
+	    long long tempLL;
+	    unsigned long long tempSize;
+
+	    NSLog(@"recognized a 150 response, looking for size in: %@", s);
+	    leftParR = [s rangeOfString:@"(" options:NSBackwardsSearch];
+	    if (leftParR.location > 0)
+	      {
+		sizeString = [s substringWithRange:NSMakeRange(leftParR.location+1, bytesR.location-leftParR.location)];
+	        [[NSScanner scannerWithString: sizeString] scanLongLong:&tempLL];
+		NSLog(@"parsed response size from %@ is %lld", sizeString, tempLL);
+		if (tempLL > 0)
+		  {
+		    tempSize = (unsigned long long)tempLL;
+		    if (fileSize == 0)
+		      fileSize = tempSize;
+		    else if (fileSize != tempSize)
+		      NSLog(@"Apparently the server is lying! list size is: %llu, transfer size is: %lld", fileSize, tempSize);
+		  }
+	      }
+	  }
+      }
     [reply release];
     
     if ([self initDataStream] < 0)
     {
         [controller showAlertDialog:@"Unexpected connection error."];
-        return;
+        return NO;
     }
     
     localFileStream = fopen([localPath cString], "w");
@@ -491,7 +526,7 @@ int getChar(streamStruct* ss)
     {
         [controller showAlertDialog:@"Opening of local file failed.\nCheck permissions and free space."];
         perror("local fopen failed");
-        return;
+        return NO;
     }
     
     totalBytes = 0;
@@ -531,69 +566,63 @@ int getChar(streamStruct* ss)
     [self readReply:&reply];
     [reply release];
     [controller setThreadRunningState:NO];
+  return gotFile;
 }
 
-- (oneway void)storeFile:(fileElement *)file from:(LocalClient *)localClient beingAt:(int)depth
+- (BOOL)storeFile:(FileElement *)file from:(LocalClient *)localClient beingAt:(int)depth
 {
-    NSString           *fileName;
-    unsigned long long fileSize;
-    char               fNameCStr[MAX_CONTROL_BUFF];
-    char               command[MAX_CONTROL_BUFF];
-    char               buff[MAX_DATA_BUFF];
-    FILE               *localFileStream;
-    NSMutableArray     *reply;
-    int                bytesRead;
-    struct sockaddr    from;
-    int                fromLen;
-    unsigned int       minimumPercentIncrement;
-    unsigned int       progressIncBytes;
-    int                replyCode;
-    unsigned long long totalBytes;
-    NSString           *localPath;
-    BOOL               gotFile;
+  NSString           *fileName;
+  unsigned long long fileSize;
+  NSString           *command;
+  char               buff[MAX_DATA_BUFF];
+  FILE               *localFileStream;
+  NSMutableArray     *reply;
+  int                bytesRead;
+  unsigned int       minimumPercentIncrement;
+  unsigned int       progressIncBytes;
+  int                replyCode;
+  unsigned long long totalBytes;
+  NSString           *localPath;
+  BOOL               gotFile;
 
-    
-    fromLen = sizeof(from);
+  fileName = [file name];
+  fileSize = [file size];
+  minimumPercentIncrement = fileSize / 100; // we should guard against maxint
 
-    fileName = [file filename];
-    fileSize = [file size];
-    minimumPercentIncrement = fileSize / 100; // we should guard against maxint
+  localPath = [file path];
 
-    localPath = [[localClient workingDir] stringByAppendingPathComponent:fileName];
-
-    if ([file isDir])
+  if ([file isDir])
     {
-        NSString     *pristineLocalPath;  /* original path */
-        NSString     *pristineRemotePath; /* original path */
-        NSArray      *dirList;
-        NSString     *remotePath;
-        NSEnumerator *en;
-        fileElement  *fEl;
+      NSString     *pristineLocalPath;  /* original path */
+      NSString     *pristineRemotePath; /* original path */
+      NSArray      *dirList;
+      NSString     *remotePath;
+      NSEnumerator *en;
+      FileElement  *fEl;
 
-        if (depth > MAX_DIR_RECURSION)
+      if (depth > MAX_DIR_RECURSION)
         {
-            NSLog(@"Max depth reached: %d", depth);
-            return;
+          NSLog(@"Max depth reached: %d", depth);
+          return NO;
         }
 
-        pristineLocalPath = [[localClient workingDir] retain];
-        pristineRemotePath = [[self workingDir] retain];
-
-        NSLog(@"it is a dir: %@", fileName);
-        remotePath = [pristineRemotePath stringByAppendingPathComponent:fileName];
-        [localClient changeWorkingDir:localPath];
-        NSLog(@"local dir changed: %@", [localClient workingDir]);
-
-        if ([self createNewDir:remotePath] == YES)
+      pristineLocalPath = [[localClient workingDir] retain];
+      pristineRemotePath = [[self workingDir] retain];
+      
+      remotePath = [pristineRemotePath stringByAppendingPathComponent:fileName];
+      [localClient changeWorkingDir:localPath];
+      NSLog(@"local dir changed: %@", [localClient workingDir]);
+      
+      if ([self createNewDir:remotePath] == YES)
         {
-            NSLog(@"remote dir created succesfully");
+            NSLog(@"remote dir created successfully");
             [self changeWorkingDir:remotePath];
 
             dirList = [localClient dirContents];
             en = [dirList objectEnumerator];
             while ((fEl = [en nextObject]))
             {
-                NSLog(@"recurse, upload : %@", [fEl filename]);
+                NSLog(@"recurse, upload : %@", [fEl name]);
                 [self storeFile:fEl from:localClient beingAt:(depth+1)];
             }
         }
@@ -602,7 +631,7 @@ int getChar(streamStruct* ss)
         [localClient changeWorkingDir:pristineLocalPath];
         [pristineLocalPath release];
         [pristineRemotePath release];
-        return;
+        return YES;
     }
     
     /* lets settle to a plain binary standard type */
@@ -612,11 +641,10 @@ int getChar(streamStruct* ss)
     {
         [controller showAlertDialog:@"Error initiating the Data Connection."];
         NSLog(@"error initiating data connection, storeFile");
-        return;
+        return NO;
     }
 
-    [fileName getCString:fNameCStr];
-    sprintf(command, "STOR %s\r\n", fNameCStr);
+    command = [@"STOR " stringByAppendingString:fileName];
     [self writeLine:command];
     replyCode = [self readReply:&reply];
     NSLog(@"%d reply is %@: ", replyCode, [reply objectAtIndex:0]);
@@ -626,7 +654,7 @@ int getChar(streamStruct* ss)
         [controller showAlertDialog:[reply objectAtIndex:0]];
         [self logIt: [reply objectAtIndex:0]];
         [reply release];
-        return;
+        return NO;
     }
     [reply release];
 
@@ -634,7 +662,7 @@ int getChar(streamStruct* ss)
     if ([self initDataStream] < 0)
     {
         [controller showAlertDialog:@"Unexpected connection error."];
-        return;
+        return NO;
     }
 
 
@@ -643,7 +671,7 @@ int getChar(streamStruct* ss)
     {
         [controller showAlertDialog:@"Opening of local file failed.\n Check permissions."];
         perror("local fopen failed");
-        return;
+        return NO;
     }
 
     totalBytes = 0;
@@ -683,28 +711,25 @@ int getChar(streamStruct* ss)
     [self readReply:&reply];
     [reply release];
     [controller setThreadRunningState:NO];
+  return gotFile;
 }
 
-- (void)deleteFile:(fileElement *)file beingAt:(int)depth
+- (void)deleteFile:(FileElement *)file beingAt:(int)depth
 {
-    NSString           *fileName;
-    NSString           *localPath;
-    NSFileManager      *fm;
-    char               command[MAX_CONTROL_BUFF];
-    NSMutableArray     *reply;
-    int                replyCode;
-
-    fm = [NSFileManager defaultManager];
-    fileName = [file filename];
-    localPath = [[self workingDir] stringByAppendingPathComponent:fileName];
-
-    if ([file isDir])
+  NSString           *fileName;
+  NSString           *command;
+  NSMutableArray     *reply;
+  int                replyCode;
+  
+  fileName = [file name];
+  
+  if ([file isDir])
     {
-        NSString     *pristineRemotePath; /* original path */
-        NSArray      *dirList;
-        NSString     *remotePath;
-        NSEnumerator *en;
-        fileElement  *fEl;
+      NSString     *pristineRemotePath; /* original path */
+      NSArray      *dirList;
+      NSString     *remotePath;
+      NSEnumerator *en;
+      FileElement  *fEl;
 
         if (depth > 3)
         {
@@ -714,17 +739,16 @@ int getChar(streamStruct* ss)
 
         pristineRemotePath = [[self workingDir] retain];
 
-        NSLog(@"it is a dir: %@", fileName);
         remotePath = [pristineRemotePath stringByAppendingPathComponent:fileName];
 
-        NSLog(@"remote dir created succesfully");
+        NSLog(@"remote dir created successfully");
         [self changeWorkingDir:remotePath];
 
         dirList = [self dirContents];
         en = [dirList objectEnumerator];
         while ((fEl = [en nextObject]))
         {
-            NSLog(@"recurse, delete : %@", [fEl filename]);
+            NSLog(@"recurse, delete : %@", [fEl name]);
             [self deleteFile:fEl beingAt:(depth+1)];
         }
 
@@ -733,7 +757,7 @@ int getChar(streamStruct* ss)
         [pristineRemotePath release];
     }
 
-    sprintf(command, "DELE %s\r\n", [fileName cString]);
+    command = [@"DELE " stringByAppendingString:fileName];
     [self writeLine:command];
     replyCode = [self readReply:&reply];
     NSLog(@"%d reply is %@: ", replyCode, [reply objectAtIndex:0]);
@@ -746,7 +770,6 @@ int getChar(streamStruct* ss)
 - (int)connect:(int)port :(char *)server
 {
     struct hostent      *hostentPtr;
-    char                *tempStr;
     socklentype         addrLen; /* socklen_t on some systems? */
     NSMutableArray      *reply;
 
@@ -761,8 +784,6 @@ int getChar(streamStruct* ss)
     remoteSockName.sin_family = PF_INET;
     remoteSockName.sin_port = htons(port);
 
-    tempStr = inet_ntoa(remoteSockName.sin_addr);
-
     if ((controlSocket = socket(PF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
     {
         perror("socket failed: ");
@@ -784,28 +805,32 @@ int getChar(streamStruct* ss)
     }
     
     initStream(&ctrlStream, controlSocket);
-    [self readReply :&reply];
+    if([self readReply :&reply] < 0)
+      return ERR_READ_FAIL;
     [reply release];
     return 0;
 }
 
 - (void)disconnect
 {
-    NSMutableArray *reply;
+  NSMutableArray *reply;
     
-    [self writeLine:"QUIT\r\n"];
-    [self readReply:&reply];
-    connected = NO;
+  [self writeLine:@"QUIT"];
+  [self readReply:&reply];
+  connected = NO;
+  closesocket(controlSocket);
 }
 
-- (int)authenticate:(char *)user :(char *)pass
+- (int)authenticate:(NSString *)user :(NSString *)pass
 {
-    char           tempStr[MAX_CONTROL_BUFF];
+    NSString       *tempStr;
     NSMutableArray *reply;
     int            replyCode;
 
-    sprintf(tempStr, "USER %s\r\n", user);
-    [self writeLine:tempStr];
+
+    tempStr = [@"USER " stringByAppendingString: user];
+    if ([self writeLine:tempStr] < 0)
+      return -2; /* we couldn't write */
     replyCode = [self readReply:&reply];
     if (replyCode == 530)
     {
@@ -816,8 +841,9 @@ int getChar(streamStruct* ss)
     }
     [reply release];
     
-    sprintf(tempStr, "PASS %s\r\n", pass);
-    [self writeLine:tempStr byLoggingIt:NO];
+    tempStr = [@"PASS " stringByAppendingString: pass];
+    if ([self writeLine:tempStr byLoggingIt:NO] < 0)
+      return -2; /* we couldn't write */
     replyCode = [self readReply:&reply];
     if (replyCode == 530)
     {
@@ -831,7 +857,7 @@ int getChar(streamStruct* ss)
     connected = YES;
 
     /* get home directory as dir we first connected to */
-    [self writeLine:"PWD\r\n"];
+    [self writeLine:@"PWD"];
     [self readReply:&reply];
     if ([reply count] >= 1)
     {
@@ -842,7 +868,7 @@ int getChar(streamStruct* ss)
         unsigned int i;
         
         line = [reply objectAtIndex:0];
-        NSLog(@"pwd reply is: %@", line);
+
         length = [line length];
         i = 0;
         while (i < length && ([line characterAtIndex:i] != '\"'))
@@ -886,7 +912,7 @@ int getChar(streamStruct* ss)
             return -1;
         }
 
-        [self writeLine:"PASV\r\n"];
+        [self writeLine:@"PASV"];
         replyCode = [self readReply:&reply];
         if (replyCode != 227)
         {
@@ -1001,12 +1027,13 @@ int getChar(streamStruct* ss)
 
     if (usesPorts == YES)
     {
-        union addrAccess { /* we use this union to extract the 8 bytes of an address */
+        union addrAccess /* we use this union to extract the 8 bytes of an address */
+        {
             struct in_addr   sinAddr;
             unsigned char    ipv4[4];
         } addr;
         NSMutableArray *reply;
-        char           tempStr[256];
+        NSString       *tempStr;
         unsigned char  p1, p2;
         int            returnCode;
         unsigned int   port;
@@ -1016,13 +1043,18 @@ int getChar(streamStruct* ss)
         port = ntohs(dataSockName.sin_port);
         p1 = (port & 0xFF00) >> 8;
         p2 = port & 0x00FF;
-        sprintf(tempStr, "PORT %u,%u,%u,%u,%u,%u\r\n", addr.ipv4[0], addr.ipv4[1], addr.ipv4[2], addr.ipv4[3], p1, p2);
+        tempStr = [NSString stringWithFormat:@"PORT %u,%u,%u,%u,%u,%u", addr.ipv4[0], addr.ipv4[1], addr.ipv4[2], addr.ipv4[3], p1, p2];
         [self writeLine:tempStr];
-        NSLog(@"port str: %s", tempStr);
+        NSLog(@"port str: %@", tempStr);
         if ((returnCode = [self readReply:&reply]) != 200)
         {
-            NSLog(@"error occoured in port command: %@", [reply objectAtIndex:0]);
-            return -1;
+          if(reply && [reply count] > 0)
+            {
+              NSLog(@"error occoured in port command: %@", [reply objectAtIndex:0]);
+              return -1;
+            }
+          NSLog(@"error in port command, no code");
+          return -2;
         }
     }
     return 0;
@@ -1080,32 +1112,30 @@ int getChar(streamStruct* ss)
  */
 - (BOOL)createNewDir:(NSString *)dir
 {
-    NSString       *remotePath;
-    char           command[MAX_CONTROL_BUFF];
-    char           pathCStr[MAX_CONTROL_BUFF];
-    NSMutableArray *reply;
-    int            replyCode;
+  NSString       *remotePath;
+  NSString       *command;
+  NSMutableArray *reply;
+  int            replyCode;
 
-    if ([dir hasPrefix:@"/"])
+  if ([dir hasPrefix:@"/"])
     {
-        NSLog(@"%@ is an absolute path", dir);
-        remotePath = dir;
+      NSLog(@"%@ is an absolute path", dir);
+      remotePath = dir;
     } else
     {
-        NSLog(@"%@ is a relative path", dir);
-        remotePath = [[self workingDir] stringByAppendingPathComponent:dir];
+      NSLog(@"%@ is a relative path", dir);
+      remotePath = [[self workingDir] stringByAppendingPathComponent:dir];
     }
 
-    [remotePath getCString:pathCStr];
-    sprintf(command, "MKD %s\r\n", pathCStr);
-    [self writeLine:command];
-    replyCode = [self readReply:&reply];
-    if (replyCode == 257)
-        return YES;
-    else
+  command =[@"MKD " stringByAppendingString:remotePath];
+  [self writeLine:command];
+  replyCode = [self readReply:&reply];
+  if (replyCode == 257)
+    return YES;
+  else
     {
-        NSLog(@"remote mkdir code: %d %@", replyCode, [reply objectAtIndex:0]);
-        return NO;
+      NSLog(@"remote mkdir code: %d %@", replyCode, [reply objectAtIndex:0]);
+      return NO;
     }
 }
 
@@ -1115,14 +1145,15 @@ int getChar(streamStruct* ss)
 {
     int                ch;
     char               buff[MAX_DATA_BUFF];
-    int                readBytes;
+    unsigned           readBytes;
     enum               states_m1 { READ, GOTR };
     enum               states_m1 state;
     NSMutableArray     *listArr;
-    fileElement        *aFile;
+    FileElement        *aFile;
     char               path[4096];
     NSMutableArray     *reply;
     int                replyCode;
+    unsigned long long transferSize;
     
     if (!connected)
         return nil;
@@ -1141,33 +1172,39 @@ int getChar(streamStruct* ss)
     listArr = [NSMutableArray arrayWithCapacity:5];
     
     [self initDataConn];
-    [self writeLine:"LIST\r\n"];
-    [self readReply:&reply];
+    [self writeLine:@"LIST"];
+    replyCode = [self readReply:&reply];
 
     if ([self initDataStream] < 0)
         return nil;
 
+    transferSize = 0;
+    [controller setTransferBegin:@"Listing" :transferSize];
     /* read the directory listing, each line being CR-LF terminated */
     state = READ;
     readBytes = 0;
     while ((ch = getChar(&dataStream)) != EOF)
-    {
+      {
         if (ch == '\r')
             state = GOTR;
         else if (ch == '\n' && state == GOTR)
-        {
-            
+          { 
             buff[readBytes] = '\0';
-            fprintf(stderr, "%s\n", buff);
             [self logIt:[NSString stringWithCString:buff]];
             state = READ; /* reset the state for a new line */
+	    transferSize += readBytes;
             readBytes = 0;
-            aFile = [[fileElement alloc] initWithLsLine:buff];
+            aFile = [[FileElement alloc] initWithLsLine:buff];
             if (aFile)
+              {
                 [listArr addObject:aFile];
-        } else
+                [aFile release];
+              }
+	    [controller setTransferProgress:[NSNumber numberWithUnsignedLongLong:transferSize]];
+          }
+        else
             buff[readBytes++] = ch;
-    }
+      }
 /* FIXME ***********    if (ferror(dataStream))
     {
         perror("error in reading data stream: ");
@@ -1176,10 +1213,39 @@ int getChar(streamStruct* ss)
          fprintf(stderr, "feof\n");
     } */
     [self closeDataStream];
+    [controller setTransferEnd:[NSNumber numberWithUnsignedLongLong:transferSize]];
 
     replyCode = [self readReply:&reply];
     
     return [NSArray arrayWithArray:listArr];
 }
 
+- (BOOL)renameFile:(FileElement *)file to:(NSString *)name
+{
+  NSString       *commandStr;
+  NSMutableArray *reply;
+  int            replyCode;
+
+  commandStr = [NSString stringWithFormat:@"RNFR %@", [file name]];
+  [self writeLine:commandStr];
+  replyCode = [self readReply:&reply];
+  if (replyCode != 350)
+    {
+      NSLog(@"Error during Rename from %d", replyCode);
+      return NO;
+    }
+
+  commandStr = [NSString stringWithFormat:@"RNTO %@", name];
+  [self writeLine:commandStr];
+  replyCode = [self readReply:&reply];
+
+  if (replyCode == 250)
+    {
+      [file setName:name];
+      return YES;
+    }
+
+  return NO;
+}
+
 @end
